;;;
;;; Universal Monitor INS8070 (SC/MP III)
;;;   Copyright (C) 2021  Haruo Asano
;;;

	CPU	8070


	INCLUDE	"config.inc"

	INCLUDE	"../common.inc"

;;;
;;; ROM	area
;;;

	ORG	ROM_B
	
	NOP			; Dummy
	JMP	CSTART		; Reset

	IF ROM_B = 0x0000
	;; Fixed address area
	JMP	DUMMY		; INTA
	JMP	DUMMY		; INTB

	ORG	0x0040		; Skip vector for CALL instruction
	ENDIF

	;; Temporary (for NSC Tiny BASIC)
	ORG	0x3060
	JMP	CONOUT
	JMP	CONIN

CSTART:
	LD	SP,=STACK

	JSR	INIT

	LD	EA,=RAM_B
	ST	EA,DSADDR
	ST	EA,SADDR
	ST	EA,GADDR

	;; Opening message
	LD	P2,=OPNMSG
	JSR	STROUT

WSTART:
	LD	P2,=PROMPT
	JSR	STROUT
	JSR	GETLIN
	LD	P2,=INBUF
	JSR	SKIPSP
	JSR	UPPER
	BZ	WSTART
	LD	E,A		; backup A

	SUB	A,='D'
	BNZ	M00
	JMP	DUMP
M00:
	LD	A,E
	SUB	A,='G'
	BNZ	M01
	JMP	GO
M01:
	LD	A,E
	SUB	A,='S'
	BNZ	M02
	JMP	SETM
M02:
	LD	A,E
	SUB	A,='L'
	BNZ	M03
	JMP	LOADH
M03:
	LD	A,E
	SUB	A,='P'
	BNZ	M04
	JMP	SAVEH
M04:

ERR:
	LD	P2,=ERRMSG
	JSR	STROUT
	JMP	WSTART

;;; Dump memory

DUMP:
	LD	A,@1,P2		; INC P2
	JSR	SKIPSP
	JSR	RDHEX
	LD	A,CNT
	BNZ	DP0
	;; No arg.
	JSR	SKIPSP
	LD	A,0,P2
	BNZ	ERR
DP00:
	LD	EA,DSADDR
	ADD	EA,=128
	ST	EA,DEADDR
	BRA	DPM
	;; 1st arg. found
DP0:
	LD	EA,VAL
	ST	EA,DSADDR
	JSR	SKIPSP
	LD	A,0,P2
	SUB	A,=','
	BZ	DP1
	LD	A,0,P2
	BZ	DP00		; No 2nd arg.
	BRA	ERR
	;;
DP1:
	LD	A,@1,P2
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LD	A,CNT
	BZ	DP1E
	LD	A,0,P2
	BNZ	DP1E
	LD	EA,VAL
	ADD	EA,=1
	ST	EA,DEADDR
	BRA	DPM
DP1E:
	JMP	ERR
DPM:
	LD	EA,DSADDR
	AND	A,=0xF0
	LD	P2,EA
	LD	A,=0
	ST	A,DSTATE
DPM0:
	JSR	DPL
	JSR	CONST
	BNZ	DPM1
	LD	A,DSTATE
	SUB	A,=2
	BP	DPM01
	BRA	DPM0
DPM01:
	;; Finish
	LD	EA,DEADDR
	ST	EA,DSADDR
	JMP	WSTART
DPM1:
	;; Abort
	LD	EA,P2
	ST	EA,DSADDR
	JSR	CONIN
	JMP	WSTART

	;; Dump line
DPL:
	LD	EA,P2
	PUSH	A
	LD	A,E
	JSR	HEXOUT2
	POP	A
	JSR	HEXOUT2
	PLI	P2,=DSEP0
	JSR	STROUT
	POP	P2

	LD	P3,=INBUF
	LD	A,=16
	ST	A,CNT
DPL0:
	JSR	DPB
	DLD	A,CNT
	BNZ	DPL0

	PLI	P2,=DSEP1
	JSR	STROUT
	POP	P2

	;; Print ASCII area
	LD	P3,=INBUF
	LD	A,=16
	ST	A,CNT
DPL1:
	LD	A,@1,P3
	SUB	A,=' '
	BP	DPL3
DPL2:
	;; Unprintable
	LD	A,='.'
	BRA	DPL4
DPL3:
	LD	A,-1,P3
	SUB	A,=0x7F
	BP	DPL2
	LD	A,-1,P3
DPL4:
	JSR	CONOUT
	DLD	A,CNT
	BNZ	DPL1

	JMP	CRLF

	;; Dump byte
DPB:
	LD	A,=' '
	JSR	CONOUT
	LD	A,DSTATE
	BNZ	DPB2
	;; Dump state 0
	LD	EA,P2
	SUB	EA,DSADDR
	BNZ	DPB0
	LD	A,E
	BZ	DPB1
	;; Still 0 or 2
DPB0:
	LD	A,=' '
	JSR	CONOUT
	LD	A,=' '
	JSR	CONOUT

	LD	A,=' '
	ST	A,@1,P3

	LD	EA,P2
	ADD	EA,=1
	LD	P2,EA
	RET

	;; Found start address
DPB1:
	LD	A,=1
	ST	A,DSTATE
DPB2:
	LD	A,DSTATE
	SUB	A,=1
	BNZ	DPB0
	;; Dump state 1
	LD	A,@1,P2
	ST	A,@1,P3
	JSR	HEXOUT2

	LD	EA,P2
	SUB	EA,DEADDR
	BNZ	DPBE
	LD	A,E
	BNZ	DPBE
	;; Found end address
	LD	A,=2
	ST	A,DSTATE
DPBE:
	RET

;;; GO address

GO:
	LD	A,@1,P2		; INC P2
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LD	A,0,P2
	BZ	GO0
	JMP	ERR
GO0:
	LD	A,CNT
	BZ	G0
	LD	EA,VAL
	ST	EA,GADDR
G0:
	LD	EA,GADDR
	SUB	EA,=1
	LD	PC,EA

;;; Set memory

SETM:
	LD	A,@1,P2		; INC P2
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LD	A,0,P2
	BZ	SM0
	JMP	ERR
SM0:
	LD	A,CNT
	BZ	SM1
	LD	EA,VAL
	ST	EA,SADDR
SM1:
	LD	A,SADDR+1
	JSR	HEXOUT2
	LD	A,SADDR
	JSR	HEXOUT2

	LD	P2,=DSEP1
	JSR	STROUT

	LD	EA,SADDR
	LD	P2,EA
	LD	A,0,P2
	JSR	HEXOUT2
	LD	A,=' '
	JSR	CONOUT

	JSR	GETLIN
	LD	P2,=INBUF
	JSR	SKIPSP
	LD	A,0,P2
	LD	E,A
	BNZ	SM2
	;; Empty (Increment address)
	LD	EA,SADDR
	ADD	EA,=1
	ST	EA,SADDR
	BRA	SM1
SM2:
	LD	A,E
	SUB	A,='-'
	BNZ	SM3
	;; '-' (Decrement address)
	LD	EA,SADDR
	SUB	EA,=1
	ST	EA,SADDR
	BRA	SM1
SM3:
	LD	A,E
	SUB	A,='.'
	BNZ	SM4
	;; '.' (Quit)
	JMP	WSTART
SM4:
	JSR	RDHEX
	LD	A,CNT
	BNZ	SM5
	JMP	ERR
SM5:
	LD	EA,SADDR
	LD	P2,EA
	ADD	EA,=1
	ST	EA,SADDR
	LD	A,VAL		; Low byte
	ST	A,0,P2
	BRA	SM1

;;; LOAD HEX file

LOADH:
	LD	A,@1,P2
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LD	A,0,P2
	BNZ	LHE

LH0:	JSR	CONIN
	JSR	UPPER
	LD	E,A
	SUB	A,='S'
	BZ	LHS0
LH1:
	LD	A,E
	SUB	A,=':'
	BZ	LHI0
LH2:
	;; Skip to EOL
	LD	A,E
	SUB	A,=CR
	BZ	LH0
	LD	A,E
	SUB	A,=LF
	BZ	LH0
LH3:
	JSR	CONIN
	LD	E,A
	BRA	LH2
LHE:
	JMP	ERR

LHI0:
	JSR	HEXIN
	ST	A,SUM
	ST	A,CNT

	JSR	HEXIN
	LD	E,A		; Address H
	ADD	A,SUM
	ST	A,SUM

	JSR	HEXIN
	PUSH	A
	ADD	A,SUM
	ST	A,SUM
	POP	A

	;; Add offset
	ADD	EA,VAL
	LD	P2,EA

	JSR	HEXIN
	ST	A,RECTYP
	ADD	A,SUM
	ST	A,SUM
	LD	A,RECTYP
	SUB	A,=2
	BP	LH3		; Skip unsupported record type

	LD	A,CNT
	BZ	LHI3
LHI1:
	JSR	HEXIN
	LD	E,A
	ADD	A,SUM
	ST	A,SUM

	LD	A,RECTYP
	BNZ	LHI2

	LD	A,E
	ST	A,@1,P2
LHI2:
	DLD	A,CNT
	BNZ	LHI1
LHI3:
	JSR	HEXIN
	ADD	A,SUM
	BNZ	LHIE		; Checksum error
	LD	A,RECTYP
	BZ	LH3
	JMP	WSTART
LHIE:
	LD	P2,=IHEMSG
	JSR	STROUT
	JMP	WSTART

LHS0:
	LD	A,=0
	ST	A,RECTYP
	JSR	CONIN
	LD	E,A
	SUB	A,='1'
	BZ	LHS02
	LD	A,E
	SUB	A,='9'
	BZ	LHS01
	BRA	LH3		; Unsupported record type
LHS01:
	LD	A,=1
	ST	A,RECTYP
LHS02:

	JSR	HEXIN
	ST	A,SUM
	SUB	A,=3
	ST	A,CNT

	JSR	HEXIN
	LD	E,A		; Address H
	ADD	A,SUM
	ST	A,SUM

	JSR	HEXIN
	PUSH	A
	ADD	A,SUM
	ST	A,SUM
	POP	A		; Address L

	;; Add offset
	ADD	EA,VAL
	LD	P2,EA

	LD	A,CNT
	BZ	LHS3
LHS1:
	JSR	HEXIN
	LD	E,A
	ADD	A,SUM
	ST	A,SUM

	LD	A,RECTYP
	BNZ	LHS2

	LD	A,E
	ST	A,@1,P2
LHS2:
	DLD	A,CNT
	BNZ	LHS1
LHS3:

	JSR	HEXIN
	ADD	A,SUM
	SUB	A,=0xFF
	BNZ	LHSE		; Checksum error

	LD	A,RECTYP
	BNZ	LHSR
	JMP	LH3
LHSE:
	LD	P2,=SHEMSG
	JSR	STROUT
LHSR:
	JMP	WSTART
	
SAVEH:

;;; Other support routines

STROUT:
	LD	A,@+1,P2
	BZ	STROR
	JSR	CONOUT
	BRA	STROUT
STROR:
	RET

HEXOUT2:
	PUSH	A
	SR	A
	SR	A
	SR	A
	SR	A
	JSR	HEXOUT1
	POP	A
HEXOUT1:
	AND	A,=0x0F
	ADD	A,='0'
	LD	E,A
	SUB	A,='9'+1
	BP	HEXOUT0
	LD	A,E
HEXOUTE:
	JSR	CONOUT
	RET
HEXOUT0:
	LD	A,E
	ADD	A,='A'-'9'-1
	BRA	HEXOUTE

HEXIN:
	LD	A,=0
	JSR	HI0
	SL	A
	SL	A
	SL	A
	SL	A
HI0:
	PUSH	EA
	JSR	CONIN
	JSR	UPPER
	LD	E,A
	SUB	A,='F'+1
	BP	HIR
	LD	A,E
	SUB	A,='A'
	BP	HI2
	LD	A,E
	SUB	A,='9'+1
	BP	HIR
	LD	A,E
	SUB	A,='0'
	BP	HI1
HIR:
	POP	EA
	RET
HI1:
	LD	A,E
	SUB	A,='0'
	BRA	HI3
HI2:
	LD	A,E
	SUB	A,='A'-10
HI3:
	OR	A,0,SP
	ST	A,0,SP
	BRA	HIR
	
CRLF:
	LD	A,=CR
	JSR	CONOUT
	LD	A,=LF
	JMP	CONOUT

GETLIN:
	LD	P2,=INBUF
	LD	A,=0
	ST	A,CNT
GL0:
	JSR	CONIN
	LD	E,A
	SUB	A,=CR
	BZ	GLE
	LD	A,E
	SUB	A,=LF
	BNZ	GL1
GLE:
	JSR	CRLF
	LD	A,=0
	ST	A,0,P2
	RET
GL1:	
	LD	A,E
	SUB	A,=BS
	BZ	GLB
	LD	A,E
	SUB	A,=DEL
	BNZ	GL2
GLB:
	LD	A,CNT
	BZ	GL0
	SUB	A,=1
	ST	A,CNT
	LD	A,@-1,P2	; DEC P2
	LD	A,=BS
	JSR	CONOUT
	LD	A,=' '
	JSR	CONOUT
	LD	A,=BS
	JSR	CONOUT
	BRA	GL0
GL2:
	LD	A,E
	SUB	A,=' '
	BP	GL3
	BRA	GL0		; 0x00-0x1F
GL3:
	LD	A,E
	SUB	A,=0x7F
	BP	GL0		; 0x7F-0xFF
	LD	A,CNT
	SUB	A,=BUFLEN-1
	BP	GL0
	ILD	A,CNT		; INC CNT
	LD	A,E
	ST	A,@1,P2
	JSR	CONOUT
	BRA	GL0

SKIPSP:
	LD	A,@1,P2
	SUB	A,=' '
	BZ	SKIPSP
	LD	A,@-1,P2
	RET

UPPER:
	LD	E,A
	SUB	A,='a'
	BP	UP0
	BRA	UPR
UP0:
	LD	A,E
	SUB	A,='z'+1
	BP	UPR
	LD	A,E
	ADD	A,='A'-'a'
	RET
UPR:
	LD	A,E
	RET

RDHEX:
	LD	EA,=0
	ST	A,CNT
	ST	EA,VAL
RH0:
	LD	A,@1,P2
	JSR	UPPER
	LD	E,A
	SUB	A,='0'
	BP	RH01
	BRA	RHE
RH01:
	LD	A,E
	SUB	A,='9'+1
	BP	RH02
	LD	A,E
	BRA	RH1
RH02:
	LD	A,E
	SUB	A,='A'
	BP	RH03
	BRA	RHE
RH03:
	LD	A,E
	SUB	A,='F'+1
	BP	RHE
	LD	A,E
	SUB	A,='A'-'9'-1
RH1:
	SUB	A,='0'
	PUSH	A
	LD	EA,VAL
	SL	EA
	SL	EA
	SL	EA
	SL	EA
	OR	A,0,SP
	ST	EA,VAL
	POP	A		; Drop
	ILD	A,CNT
	BRA	RH0
RHE:
	LD	A,@-1,P2	; DEC P2
	RET

DUMMY:
	RET

;;; Data area

OPNMSG:
	DB	CR,LF,"Universal Monitor INS8070",CR,LF,0x00
PROMPT:
	DB	"] ",0x00

IHEMSG:
	DB	"Error ihex",CR,LF,0x00
SHEMSG:
	DB	"Error srec",CR,LF,0x00
ERRMSG:
	DB	"Error",CR,LF,0x00

DSEP0:
	DB	" :",0x00
DSEP1:
	DB	" : ",0x00


	IF USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF

;;;
;;; RAM area
;;;

	;; Work area

	ORG	WORK_B

INBUF:	DS	BUFLEN		; Line input buffer
DSADDR:	DS	2		; Dump start address
DEADDR:	DS	2		; Dump end address
DSTATE:	DS	1		; Dump state
GADDR:	DS	2		; Go address
SADDR:	DS	2		; Set address
HEXMOD:	DS	1		; HEX file mode

VAL:	DS	2		; Generic 16bit value
CNT:	DS	1		; Generic Counter
SUM:	DS	1		; Checksum
RECTYP:	DS	1		; Record type

	END
